@@ -1,5 +1,5 @@
-// Copyright 2008-present Contributors to the OpenImageIO project.
-// SPDX-License-Identifier: BSD-3-Clause
+// Copyright Contributors to the OpenImageIO project.
+// SPDX-License-Identifier: Apache-2.0
 // https://github.com/OpenImageIO/oiio
 
 
@@ -196,8 +196,10 @@ public:
         m_chars = make_unique(sref);
     }
 
+#ifndef __CUDA_ARCH__
     /// Construct from a known ustringhash
     inline explicit ustring(ustringhash hash);
+#endif
 
     /// ustring destructor.
     ~ustring() noexcept {}
@@ -684,9 +686,7 @@ public:
     }
 
     /// Concatenate two strings, returning a ustring, implemented carefully
-    /// to not perform any redundant copies or allocations. This is
-    /// semantically equivalent to `ustring::sprintf("%s%s", s, t)`, but is
-    /// more efficient.
+    /// to not perform any redundant copies or allocations.
     static ustring concat(string_view s, string_view t);
 
     /// Generic stream output of a ustring.
@@ -813,26 +813,51 @@ public:
     }
 
     /// Construct a ustringhash from a null-terminated C string (char *).
-    OIIO_HOSTDEVICE explicit ustringhash(const char* str)
+    OIIO_DEVICE_CONSTEXPR explicit ustringhash(const char* str)
     {
 #ifdef __CUDA_ARCH__
-        m_hash = Strutil::strhash(str);  // GPU: just compute the hash
+        // GPU: just compute the hash. This can be constexpr!
+        m_hash = Strutil::strhash(str);
 #else
-        m_hash = ustring(str).hash();  // CPU: make ustring, get its hash
+        // CPU: make ustring, get its hash. Note that ustring ctr can't be
+        // constexpr because it has to modify the internal ustring table.
+        m_hash = ustring(str).hash();
+#endif
+    }
+
+    OIIO_DEVICE_CONSTEXPR explicit ustringhash(const char* str, size_t len)
+    {
+#ifdef __CUDA_ARCH__
+        // GPU: just compute the hash. This can be constexpr!
+        m_hash = Strutil::strhash(len, str);
+#else
+        // CPU: make ustring, get its hash. Note that ustring ctr can't be
+        // constexpr because it has to modify the internal ustring table.
+        m_hash = ustring(str, len).hash();
 #endif
     }
 
     /// Construct a ustringhash from a string_view, which can be
     /// auto-converted from either a std::string.
-    OIIO_HOSTDEVICE explicit ustringhash(string_view str)
+    OIIO_DEVICE_CONSTEXPR explicit ustringhash(string_view str)
     {
 #ifdef __CUDA_ARCH__
-        m_hash = Strutil::strhash(str);  // GPU: just compute the hash
+        // GPU: just compute the hash. This can be constexpr!
+        m_hash = Strutil::strhash(str);
 #else
-        m_hash = ustring(str).hash();  // CPU: make ustring, get its hash
+        // CPU: make ustring, get its hash. Note that ustring ctr can't be
+        // constexpr because it has to modify the internal ustring table.
+        m_hash = ustring(str).hash();
 #endif
     }
 
+    /// Construct from a raw hash value. Beware: results are undefined if it's
+    /// not the valid hash of a ustring.
+    OIIO_HOSTDEVICE explicit constexpr ustringhash(size_t hash) noexcept
+        : m_hash(hash)
+    {
+    }
+
     /// Conversion to an OIIO::string_view.
     operator string_view() const noexcept { return ustring::from_hash(m_hash); }
 
@@ -961,7 +986,7 @@ public:
 #endif
 
     /// Return the ustringhash corresponding to the given hash. Caveat emptor:
-    /// results are undefined if it's not the valud hash of a ustring.
+    /// results are undefined if it's not the hash value of a ustring.
     OIIO_NODISCARD static constexpr ustringhash from_hash(size_t hash)
     {
         ustringhash u;
@@ -973,13 +998,6 @@ private:
     // Individual ustringhash internal representation -- the hash value.
     rep_t m_hash;
 
-    // Construct from a raw hash value. It's protected so that it's only
-    // callable by its friend, ustring.
-    OIIO_HOSTDEVICE ustringhash(size_t hashval)
-        : m_hash(hashval)
-    {
-    }
-
     friend class ustring;
 };
 
@@ -1011,6 +1029,22 @@ inline ustring::ustring(ustringhash hash)
 
 
 
+/// ustring string literal operator
+inline ustring operator""_us(const char* str, std::size_t len)
+{
+    return ustring(str, len);
+}
+
+
+/// ustringhash string literal operator
+OIIO_DEVICE_CONSTEXPR ustringhash operator""_ush(const char* str,
+                                                 std::size_t len)
+{
+    return ustringhash(str, len);
+}
+
+
+
 #if OIIO_VERSION_LESS(3, 0, 0)
 /// Deprecated -- This is too easy to confuse with the ustringhash class. And
 /// also it is unnecessary if you use std::hash<ustring>. This will be removed
@@ -1072,6 +1106,8 @@ iequals(const std::string& a, ustring b)
 // ustring variant stof from OpenImageIO/strutil.h
 namespace Strutil {
 
+#ifndef __CUDA_ARCH__
+
 inline float
 stof(ustring s)
 {
@@ -1092,6 +1128,8 @@ to_string(const ustringhash& value)
     return ustring(value).string();
 }
 
+#endif
+
 }  // end namespace Strutil
 
 OIIO_NAMESPACE_END
